unit ControllerU;

interface

uses Graphics, Contnrs, Windows, GenAlgU, S2DVectorU, Matrix2DU, UtilsU,
     VectorU;

Type
   TController = class(TObject)
   private
      FSweepers: TObjectList;
      FMines: TSVector2DList;

      FGA: TGeneticAlgorithm;

      FNumSweepers: Integer;
      FNumMines: Integer;
      FNumWeightsInNN: Integer;

      FAvFitness: TVector;
      FBestFitness: TVector;

      FFastRender: Boolean;
      FTicks: Integer;
      FGenerations: Integer;

      FcxClient, FcyClient: Integer;

      procedure PlotStats(ACanvas: TCanvas);
   public
      Constructor Create;
      Destructor Destroy; Override;

      procedure Render(ACanvas: TCanvas);
      procedure WorldTransform(var VBuffer: Array of TSPoint; vPos: TSVector2D);

      procedure Update;
      procedure FastRenderToggle;

      property FastRender: Boolean read FFastRender write FFastRender;
   end;

implementation

uses SysUtils, ParamsU, MineSweeperU;

Const
   NumSweeperVerts = 16;
   NumMineVerts = 4;

Type
   TSweeperVerts = Array [0..NumSweeperVerts-1] of TSPoint;
   TMineVerts = Array [0..NumMineVerts-1] of TSPoint;

Const
   SweeperVerts: TSweeperVerts =
      (
         (x:-1; y:-1), (x:-1; y:1), (x:-0.5; y:1), (x:-0.5; y:-1),

         (x:0.5; y:-1), (x:1; y:-1), (x:1; y:1), (x:0.5; y:1),

         (x:-0.5; y:-0.5), (x:0.5; y:-0.5),

         (x:-0.5; y:0.5),  (x:-0.25; y:0.5), (x:-0.25; y:1.75), (x:0.25; y:1.75),
         (x:0.25; y:0.5), (x:0.5; y:0.5)
      );

   MineVerts: TMineVerts =
      (
         (x:-1; y:-1), (x:-1; y:1), (x:1; y:1), (x:1; y:-1)
      );


{ TController }

constructor TController.Create;
var
   i: Integer;
   ASweeper: TMineSweeper;
   AMine: TSVector2D;
begin
   FAvFitness := TVector.Create;
   FBestFitness := TVector.Create;

   FNumSweepers := Params.NumSweepers;
   FGA := nil;
   FFastRender := false;
   FTicks := 0;
   FNumMines := Params.NumMines;
   FGenerations := 0;
   FcxClient := Params.WindowWidth;
   FcyClient := Params.WindowHeight;

   FSweepers := TObjectList.Create;

   For i := 0 to FNumSweepers -1 do
   begin
      ASweeper := TMineSweeper.Create;
      FSweepers.Add(ASweeper);
   end;

   FNumWeightsInNN := TMineSweeper(FSweepers[0]).WeightCount;

   FGA := TGeneticAlgorithm.Create(FNumSweepers, Params.MutationRate,
                                   Params.CrossoverRate, FNumWeightsInNN);

   For i := 0 to FNumSweepers -1 do
   begin
      ASweeper := TMineSweeper(FSweepers[i]);
      ASweeper.PutWeights(FGA.Chromos[i].Weights);
   end;

   FMines := TSVector2DList.Create;

   For i := 0 to FNumMines -1 do
   begin
      AMine := FMines.Add;
      AMine.X := RandFloat * FcxClient;
      AMine.Y := RandFloat * FcyClient;
   end;
end;

destructor TController.Destroy;
begin
   FAvFitness.Free;
   FBestFitness.Free;
   FSweepers.Free;
   FMines.Free;
   FGA.Free;

   inherited;
end;

procedure TController.FastRenderToggle;
begin
   FFastRender := not FFastRender;
end;

procedure TController.PlotStats(ACanvas: TCanvas);
var
   s: String;
   HSlice, VSlice, x: Double;
   i: Integer;
begin
   With ACanvas do
   begin
      s := 'Best Fitness:        ' + FloatToStr(FGA.BestFitness);
      TextOut(5, 20, s);

      s := 'Average Fitness:     ' + FloatToStr(FGA.AverageFitness);
      TextOut(5, 40, s);

      HSlice := FcxClient / (FGenerations + 1);
      VSlice := FcyClient / ((FGA.BestFitness +1) * 2);

      x := 0;
      Pen.Color := clRed;

      MoveTo(0, FcyClient);
      For i := 0 to FBestFitness.Count -1 do
      begin
         LineTo(Round(x), Round(FcyClient - VSlice*FBestFitness[i]));
         x := x + HSlice;
      end;

      x := 0;
      Pen.Color := clBlue;

      MoveTo(0, FcyClient);

      For i := 0 to FAvFitness.Count -1 do
      begin
         LineTo(Round(x), Round(FcyClient - VSlice*FAvFitness[i]));
         x := x + HSlice;
      end;
   end;
end;

procedure TController.Render(ACanvas: TCanvas);
var
   s: String;
   i, Vert: Integer;
   ASweeper: TMineSweeper;
   asv: TSweeperVerts;
   amv: TMineVerts;
begin
   With ACanvas do
   begin
      Font.Style := [fsBold];
      s := 'Generation:          ' + IntToStr(FGenerations);
      ACanvas.TextOut(5, 0, s);

      if not FFastRender then
      begin
         Pen.Color := clGreen;

         For i := 0 to FNumMines -1 do
         begin
            amv := MineVerts;
            WorldTransform(amv, FMines[i]);
            MoveTo(Round(amv[0].X), Round(amv[0].Y));

            For Vert := 1 to NumMineVerts -1 do
            begin
               LineTo(Round(amv[Vert].x), Round(amv[Vert].y));
            end;
            LineTo(Round(amv[0].X), Round(amv[0].Y));
         end;

         Pen.Color := clRed;

         For i := 0 to FNumSweepers -1 do
         begin
            ASweeper := TMineSweeper(FSweepers[i]);
            if i = Params.NumElite then
               Pen.Color := clBlack;

            asv := SweeperVerts;
            ASweeper.WorldTransform(asv);

            // Left Track
            MoveTo(Round(asv[0].x), Round(asv[0].y));
            For vert := 1 to 3 do
            begin
               LineTo(Round(asv[vert].x), Round(asv[vert].y));
            end;

            LineTo(Round(asv[0].x), Round(asv[0].y));

            // Right Track
            MoveTo(Round(asv[4].x), Round(asv[4].y));
            For vert := 5 to 7 do
            begin
               LineTo(Round(asv[vert].x), Round(asv[vert].y));
            end;

            LineTo(Round(asv[4].x), Round(asv[4].y));

            MoveTo(Round(asv[8].x), Round(asv[8].y));
            LineTo(Round(asv[9].x), Round(asv[9].y));

            MoveTo(Round(asv[10].x), Round(asv[10].y));
            For vert := 11 to 15 do
            begin
               LineTo(Round(asv[vert].x), Round(asv[vert].y));
            end;
         end;
      end
      else
         PlotStats(ACanvas);
   end;
end;

procedure TController.Update;
var
   i: Integer;
   ASweeper: TMineSweeper;
   GrabHit: Integer;
begin
   if FTicks < Params.NumTicks then
   begin
      For i := 0 to FNumSweepers -1 do
      begin
         ASweeper := TMineSweeper(FSweepers[i]);
         if not ASweeper.Update(FMines) then
         begin
            Raise Exception.Create('Wrong amount of NN inputs');
         end;

         GrabHit := ASweeper.CheckForMine(FMines, Params.MineScale);

         if GrabHit >= 0 then
         begin
            ASweeper.IncrementFitness;
            FMines[GrabHit].X := RandFloat * FcxClient;
            FMines[GrabHit].Y := RandFloat * FcyClient;

            FGA.Chromos[i].Fitness := ASweeper.Fitness;
         end;
      end;

      Inc(FTicks);
   end
   else
   begin
      FAvFitness.Add(FGA.AverageFitness);
      FBestFitness.Add(FGA.BestFitness);

      Inc(FGenerations);
      FTicks := 0;

      FGA.Epoch;

      For i := 0 to FNumSweepers -1 do
      begin
         ASweeper := TMineSweeper(FSweepers[i]);
         ASweeper.PutWeights(FGA.Chromos[i].Weights);
         ASweeper.Reset;
      end;
   end;
end;

procedure TController.WorldTransform(var VBuffer: Array of TSPoint;
  vPos: TSVector2D);
var
   matTransform: T2DMatrix;
begin
   matTransform := T2DMatrix.Create;
   try
      matTransform.Scale(Params.MineScale, Params.MineScale);

      matTransform.Translate(vPos.X, vPos.Y);

      matTransform.TransFormPoints(VBuffer);
   finally
      matTransform.Free;
   end;
end;

end.
