unit GenAlgU;

interface

uses Contnrs, VectorU;

Type
   TGenome = class(TObject)
   private
      FWeights: TVector;
      FFitness: Double;

   public
      Constructor Create; Overload;
      Constructor Create(w: TVector; f: Double); Overload;
      Destructor Destroy; Override;

      procedure Assign(AGenome: TGenome);

      property Weights: TVector read FWeights;
      property Fitness: Double read FFitness write FFitness;
   end;

   TPopulation = class(TObject)
   private
      FGenomes: TObjectList;
      function GetCount: Integer;
      function GetGenomes(Idx: Integer): TGenome;

   public
      Constructor Create;
      Destructor Destroy; Override;

      function Add: TGenome; Overload;
      function Add(AWeights: TVector; AFitness: Double): TGenome; Overload;
      procedure Clear;

      procedure Assign(APop: TPopulation);
      procedure Sort;

      property Count: Integer read GetCount;
      property Genomes[Idx: Integer]: TGenome read GetGenomes; Default;
   end;

   TGeneticAlgorithm = class(TObject)
   private
      FPopulation: TPopulation;
      FPopSize: Integer;
      FChromoLength: Integer;
      FTotalFitness: Double;
      FBestFitness: Double;
      FAverageFitness: Double;
      FWorstFitness: Double;
      FFittestGenome: Integer;
      FMutationRate: Double;
      FCrossoverRate: Double;

      FGenerationCount: Integer;

      procedure Crossover(Mom, Dad, Baby1, Baby2: TVector);
      procedure Mutate(Chromo: TVector);
      function  GetChromoRoulette: TGenome;
      procedure GrabNBest(NBest, NumCopies: Integer; Population: TPopulation);
      procedure CalculateBestWorstAvTot;
      procedure Reset;
      function GetAverageFitness: Double;
   public
      Constructor Create(PopSize: Integer; MutRat, CrossRat: Double; NumWeights: Integer);
      Destructor Destroy; Override;

      procedure Epoch;

      property AverageFitness: Double read GetAverageFitness;
      property BestFitness: Double read FBestFitness;
      property Chromos: TPopulation read FPopulation;
   end;

implementation

uses UtilsU, ParamsU;

function GenomeComp(Gene1, Gene2: TGenome): Integer;
begin
   Result := 0;
   if Gene1.Fitness < Gene2.Fitness then
      Result := -1
   else if Gene1.Fitness > Gene2.Fitness then
      Result := 1;
end;

{ TGenome }

constructor TGenome.Create;
begin
   FWeights := TVector.Create;
   FFitness := 0;
end;

procedure TGenome.Assign(AGenome: TGenome);
begin
   FWeights.Assign(AGenome.Weights);
   FFitness := AGenome.Fitness;
end;

constructor TGenome.Create(w: TVector; f: Double);
begin
   FWeights := TVector.Create;
   FWeights.Assign(w);
   FFitness := f;
end;

destructor TGenome.Destroy;
begin
   FWeights.Free;
   inherited;
end;

{ TPopulation }

function TPopulation.Add: TGenome;
begin
   Result := TGenome.Create;
   FGenomes.Add(Result);
end;

function TPopulation.Add(AWeights: TVector; AFitness: Double): TGenome;
var
   AGenome: TGenome;
begin
   AGenome := Add;
   AGenome.Weights.Assign(AWeights);
   AGenome.Fitness := AFitness;
   Result := AGenome;
end;

procedure TPopulation.Assign(APop: TPopulation);
var
   AGenome: TGenome;
   i: Integer;
begin
   Clear;
   For i := 0 to APop.Count -1 do
   begin
      AGenome := Add;
      AGenome.Assign(APop[i]);
   end;
end;

procedure TPopulation.Clear;
begin
   FGenomes.Clear;
end;

constructor TPopulation.Create;
begin
   FGenomes := TObjectList.Create;
end;

destructor TPopulation.Destroy;
begin
   FGenomes.Free;
   inherited;
end;

function TPopulation.GetCount: Integer;
begin
   Result := FGenomes.Count;
end;

function TPopulation.GetGenomes(Idx: Integer): TGenome;
begin
   Result := TGenome(FGenomes[Idx]);
end;

procedure TPopulation.Sort;
begin
   FGenomes.Sort(@GenomeComp);
end;

{ TGeneticAlgorithm }

procedure TGeneticAlgorithm.CalculateBestWorstAvTot;
var
   HighestSoFar, LowestSoFar: Double;
   i: Integer;
begin
   FTotalFitness := 0;

   HighestSoFar := 0;
   LowestSoFar := 9999999;

   For i := 0 to FPopSize -1 do
   begin
		//update fittest if necessary
      if FPopulation[i].Fitness > HighestSoFar then
      begin
         HighestSoFar := FPopulation[i].Fitness;
         FFittestGenome := i;
         FBestFitness := HighestSoFar;
      end;

      if FPopulation[i].Fitness < LowestSoFar then
      begin
         LowestSoFar := FPopulation[i].Fitness;
         FWorstFitness := LowestSoFar;
      end;

      FTotalFitness := FTotalFitness + FPopulation[i].Fitness;
   end;

   FAverageFitness := FTotalFitness / FPopSize;
end;

constructor TGeneticAlgorithm.Create(PopSize: Integer; MutRat,
  CrossRat: Double; NumWeights: Integer);
var
   i, j: Integer;
   AGenome: TGenome;
begin
   FPopulation := TPopulation.Create;
   FPopSize := PopSize;
   FMutationRate := MutRat;
   FCrossoverRate := CrossRat;
   FChromoLength := NumWeights;
   FGenerationCount := 0;
   FFittestGenome := 0;

   Reset;

   For i := 0 to FPopSize -1 do
   begin
      AGenome := FPopulation.Add;
      For j := 0 to FChromoLength -1 do
      begin
         AGenome.Weights.Add(RandClamped);
      end;
   end;
end;

procedure TGeneticAlgorithm.Crossover(Mom, Dad, Baby1, Baby2: TVector);
var
   i, cp: Integer;
begin
   if (RandFloat > FCrossoverRate) or (Mom = Dad) then
   begin
      Baby1.Assign(Mom);
      Baby2.Assign(Dad);
      Exit;
   end;

   cp := RandInt(0, FChromoLength -1);

   For i := 0 to cp -1 do
   begin
      Baby1.Add(Mom[i]);
      Baby2.Add(Dad[i]);
   end;

   For i := cp to Mom.Count -1 do
   begin
      Baby1.Add(Dad[i]);
      Baby2.Add(Mom[i]);
   end;
end;

destructor TGeneticAlgorithm.Destroy;
begin
   FPopulation.Free;
   inherited;
end;

procedure TGeneticAlgorithm.Epoch;
var
   TmpNewPop: TPopulation;
   Mom, Dad: TGenome;
   Baby1, Baby2: TVector;
begin
   Reset;

   FPopulation.Sort;

   CalculateBestWorstAvTot;

   TmpNewPop := TPopulation.Create;
   try
      if ((Params.NumCopiesElite * Params.NumElite) mod 2) = 0 then
      begin
         GrabNBest(Params.NumElite, Params.NumCopiesElite, TmpNewPop);
      end;

      While TmpNewPop.Count < FPopSize do
      begin
         Mom := GetChromoRoulette;
         Dad := GetChromoRoulette;

         Baby1 := TVector.Create;
         Baby2 := TVector.Create;

         try
            Crossover(Mom.Weights, Dad.Weights, Baby1, Baby2);

            Mutate(Baby1);
            Mutate(Baby2);

            TmpNewPop.Add(Baby1, 0);
            TmpNewPop.Add(Baby2, 0);
         finally
            Baby1.Free;
            Baby2.Free;
         end;
      end;

      FPopulation.Assign(TmpNewPop);
   finally
      TmpNewPop.Free;
   end;
end;

function TGeneticAlgorithm.GetAverageFitness: Double;
begin
   Result := FTotalFitness / FPopSize;
end;

function TGeneticAlgorithm.GetChromoRoulette: TGenome;
var
   Slice: Double;
   TheChosenOne: TGenome;
   FitnessSoFar: Double;
   i: Integer;
begin
   Slice := RandFloat * FTotalFitness;
   FitnessSoFar := 0;
   TheChosenOne := nil;

   For i := 0 to FPopSize -1 do
   begin
      FitnessSoFar := FitnessSoFar + TGenome(FPopulation[i]).Fitness;

      if FitnessSoFar >= Slice then
      begin
         TheChosenOne := TGenome(FPopulation[i]);
         break;
      end;
   end;

   Result := TheChosenOne;
end;

procedure TGeneticAlgorithm.GrabNBest(NBest, NumCopies: Integer;
  Population: TPopulation);
var
   i: Integer;
   AGenome: TGenome;
begin
   While (NBest > 0) do
   begin
      For i := 0 to NumCopies -1 do
      begin
         AGenome := Population.Add;
         AGenome.Assign(FPopulation[FPopSize -1 -NBest]);
         Dec(NBest);
      end;
   end;
end;

procedure TGeneticAlgorithm.Mutate(Chromo: TVector);
var
   i: Integer;
begin
   For i := 0 to Chromo.Count -1 do
   begin
      if RandFloat < FMutationRate then
      begin
         Chromo[i] := Chromo[i] + (RandClamped * Params.MaxPerturbation);
      end;
   end;
end;

procedure TGeneticAlgorithm.Reset;
begin
   FTotalFitness := 0;
   FBestFitness := 0;
   FWorstFitness := 9999999;
   FAverageFitness := 0;
end;

end.
