unit MinesweeperU;

interface

uses NeuralNetU, S2DVectorU, Matrix2DU, UtilsU, VectorU;

Type
   TMinesweeper = class(TObject)
   private
      FBrain: TNeuralNet;

      FPosition: TSVector2D;

      FDirection: TSVector2D;
      FRotation: Double;
      FSpeed: Double;

      FLTrack, FRTrack: Double;
      FFitness: Double;
      FScale: Double;
      FClosestMine: Integer;
      function GetWeightCount: Integer;
      procedure GetClosestMine(objects: TSVector2DList; Closest: TSVector2D);

   public
      Constructor Create;
      Destructor Destroy; Override;

      function Update(Mines: TSVector2DList): Boolean;
      procedure WorldTransform(var sweeper: Array of TSPoint);
      function CheckForMine(mines: TSVector2DList; Size: Double): Integer;
      procedure Reset;
      procedure PutWeights(w: TVector);
      procedure IncrementFitness;

      property Position: TSVector2D read FPosition;
      property Fitness: Double read FFitness;
      property WeightCount: Integer read GetWeightCount;
   end;

implementation

uses SysUtils, ParamsU;

{ TMinesweeper }

function TMinesweeper.CheckForMine(mines: TSVector2DList;
  Size: Double): Integer;
var
   DistToObject: TSVector2D;
begin
   Result := -1;

   DistToObject := TSVector2D.Create;
   try
      DistToObject.Assign(FPosition);
      DistToObject.SubVector(mines[FClosestMine]);

      if DistToObject.Length < (Size + 5) then
         Result := FClosestMine;
   finally
      DistToObject.Free;
   end;
end;

constructor TMinesweeper.Create;
begin
   FBrain := TNeuralNet.Create;
   FPosition := TSVector2D.Create;
   FDirection := TSVector2D.Create;

   Reset;

   FLTrack := 0.16;
   FRTrack := 0.16;

   FScale := Params.SweeperScale;

   FClosestMine := 0;
end;

destructor TMinesweeper.Destroy;
begin
   FPosition.Free;
   FDirection.Free;
   FBrain.Free;
   inherited;
end;

procedure TMinesweeper.GetClosestMine(objects: TSVector2DList; Closest: TSVector2D);
var
   closest_so_far, len_to_object: Double;
   vClosestObject, temp: TSVector2D;
   i: Integer;
begin
   closest_so_far := 99999;
   vClosestObject := TSVector2D.Create;
   try
      temp := TSVector2D.Create;
      try

         For i := 0 to objects.Count -1 do
         begin
            temp.Assign(Objects[i]);
            temp.SubVector(FPosition);
            len_to_object := temp.Length;

            if len_to_object < closest_so_far then
            begin
               closest_so_far := len_to_object;
               vClosestObject.Assign(FPosition);
               vClosestObject.SubVector(objects[i]);
               FClosestMine := i;
            end;
         end;
      finally
         temp.Free;
      end;

      Closest.Assign(vClosestObject);
   finally
      vClosestObject.Free;
   end;
end;

function TMinesweeper.GetWeightCount: Integer;
begin
   Result := FBrain.GetNumberOfWeights;
end;

procedure TMinesweeper.IncrementFitness;
begin
   FFitness := FFitness + 1;
end;

procedure TMinesweeper.PutWeights(w: TVector);
begin
   FBrain.PutWeights(w);
end;

procedure TMinesweeper.Reset;
begin
   FPosition.X := RandFloat * Params.WindowWidth;
   FPosition.Y := RandFloat * Params.WindowHeight;

   FFitness := 0;

   FRotation := RandFloat * Params.TwoPi;
end;

function TMinesweeper.Update(Mines: TSVector2DList): Boolean;
var
   Inputs, Outputs: TVector;
   vClosestMine, temp: TSVector2D;
   RotForce: Double;
begin
   Result := False;

   vClosestMine := TSVector2D.Create;
   try
      GetClosestMine(Mines, vClosestMine);
      vClosestMine.Normalize;

      Inputs := TVector.Create;
      Outputs := TVector.Create;
      try
         Inputs.Add(vClosestMine.X);
         Inputs.Add(vClosestMine.Y);

         Inputs.Add(FDirection.X);
         Inputs.Add(FDirection.Y);

         FBrain.Update(Inputs, Outputs);

         if Outputs.Count < Params.NumOutputs then
         begin
            raise Exception.Create('Invalid output count returned from brain.update');
         end;

         FLTrack := Outputs[0];
         FRTrack := Outputs[1];
      finally
         Inputs.Free;
         Outputs.Free;
      end;

      RotForce := FLTrack - FRTrack;

      Clamp(RotForce, -Params.MaxTurnRate, Params.MaxTurnRate);

      FRotation := FRotation + RotForce;
      FSpeed := FLTrack + FRTrack;

      FDirection.X := -Sin(FRotation);
      FDirection.Y := Cos(FRotation);

      temp := TSVector2D.Create;
      try
         temp.Assign(FDirection);
         temp.Multiply(FSpeed);

         FPosition.AddVector(temp);
      finally
         temp.Free;
      end;

      if FPosition.X > Params.WindowWidth then FPosition.X := 0;
      if FPosition.X < 0 then FPosition.X := Params.WindowWidth;
      if FPosition.Y > Params.WindowHeight then FPosition.Y := 0;
      if FPosition.Y < 0 then FPosition.Y := Params.WindowHeight;

      Result := True;
   finally
      vClosestMine.Free;
   end;
end;

procedure TMinesweeper.WorldTransform(var sweeper: Array of TSPoint);
var
   matTransform: T2DMatrix;
begin
   matTransform := T2DMatrix.Create;
   try
      matTransform.Scale(FScale, FScale);

      matTransform.Rotate(FRotation);

      matTransform.Translate(FPosition.X, FPosition.Y);

      matTransform.TransformPoints(sweeper);
   finally
      matTransform.Free;
   end;
end;

end.
