unit NeuralNetU;

interface

uses Contnrs, VectorU;

Type
   TNeuron = class(TObject)
   private
      FNumInputs: Integer;
      FWeights: TVector;
   public
      Constructor Create(ANumInputs: Integer);
      Destructor Destroy; Override;
      property NumInputs: Integer read FNumInputs;
      property Weights: TVector read FWeights;
   end;

   TNeuronLayer = class(TObject)
   private
      FNeurons: TObjectList;
      FNumNeurons: Integer;
      function GetNeurons(Idx: Integer): TNeuron;
   public
      Constructor Create(ANumNeurons, ANumInputsPerNeuron: Integer);
      Destructor Destroy; Override;

      property NumNeurons: Integer read FNumNeurons;
      property Neurons[Idx: Integer]: TNeuron read GetNeurons;
   end;

   TNeuralNet = class(TObject)
   private
      FNumInputs: Integer;
      FNumOutputs: Integer;
      FNumHideLayers: Integer;
      FNumHideNeurons: Integer;

      FLayers: TObjectList;
      procedure CreateNet;
      function Sigmoid(Activation, Response: Double): Double;
   public
      Constructor Create;
      Destructor Destroy; Override;
      procedure GetWeights(AVector: TVector);
      function GetNumberOfWeights: Integer;
      procedure PutWeights(AWeights: TVector);
      procedure Update(Inputs, Outputs: TVector);
   end;

implementation

uses SysUtils, UtilsU, ParamsU;

{ TNeuron }

constructor TNeuron.Create(ANumInputs: Integer);
var
   i: Integer;
begin
   FNumInputs := ANumInputs;
   FWeights := TVector.Create;

   // we need an additional weight for the bias, hence the lack of a -1 on FNumInputs
   For i := 0 to FNumInputs do
   begin
      FWeights.Add(RandClamped);
   end;
end;

destructor TNeuron.Destroy;
begin
   FWeights.Free;
   inherited;
end;

{ TNeuronLayer }

constructor TNeuronLayer.Create(ANumNeurons, ANumInputsPerNeuron: Integer);
var
   i: Integer;
   ANeuron: TNeuron;
begin
   FNeurons := TObjectList.Create;
   FNumNeurons := ANumNeurons;

   For i := 0 to FNumNeurons -1 do
   begin
      ANeuron := TNeuron.Create(ANumInputsPerNeuron);
      FNeurons.Add(ANeuron);
   end;
end;

destructor TNeuronLayer.Destroy;
begin
   FNeurons.Free;
   inherited;
end;

function TNeuronLayer.GetNeurons(Idx: Integer): TNeuron;
begin
   Result := TNeuron(FNeurons[Idx]);
end;

{ TNeuralNet }

constructor TNeuralNet.Create;
begin
   FLayers := TObjectList.Create;

   FNumInputs := Params.NumInputs;
   FNumOutputs := Params.NumOutputs;
   FNumHideNeurons := Params.NeuronsPerHiddenLayer;
   FNumHideLayers := Params.NumHidden;

	CreateNet;
end;

procedure TNeuralNet.CreateNet;
var
   ANeuronLayer: TNeuronLayer;
   i: Integer;
begin
	//create the layers of the network
   if FNumHideLayers > 0 then
   begin
      // Create first hidden layer
      ANeuronLayer := TNeuronLayer.Create(FNumHideNeurons, FNumInputs);
      FLayers.Add(ANeuronLayer);

      For i := 1 to FNumHideLayers -1 do
      begin
         ANeuronLayer := TNeuronLayer.Create(FNumHideNeurons, FNumHideNeurons);
         FLayers.Add(ANeuronLayer);
      end;

      //create output layer
      ANeuronLayer := TNeuronLayer.Create(FNumOutputs, FNumHideNeurons);
      FLayers.Add(ANeuronLayer);
   end
   else
   begin
      //create output layer
      ANeuronLayer := TNeuronLayer.Create(FNumOutputs, FNumInputs);
      FLayers.Add(ANeuronLayer);
   end;
end;

destructor TNeuralNet.Destroy;
begin
   FLayers.Free;
   inherited;
end;

function TNeuralNet.GetNumberOfWeights: Integer;
var
   Weights: TVector;
begin
   Weights := TVector.Create;
   try
      GetWeights(Weights);
      Result := Weights.Count;
   finally
      Weights.Free;
   end;
end;

procedure TNeuralNet.GetWeights(AVector: TVector);
var
   i, j, k: Integer;
   ALayer: TNeuronLayer;
   ANeuron: TNeuron;
   AWght: Double;
begin
   AVector.Clear;
   For i := 0 to FNumHideLayers do
   begin
      ALayer := TNeuronLayer(FLayers[i]);
      For j := 0 to ALayer.NumNeurons -1 do
      begin
         ANeuron := ALayer.Neurons[j];
         For k := 0 to ANeuron.Weights.Count -1 do
         begin
            AWght := ANeuron.Weights[k];
            AVector.Add(AWght);
         end;
      end;
   end;
end;

procedure TNeuralNet.PutWeights(AWeights: TVector);
var
   i, j, k, c: Integer;
   ALayer: TNeuronLayer;
   ANeuron: TNeuron;
   AWght: Double;
begin
   c := 0;
   For i := 0 to FNumHideLayers do
   begin
      ALayer := TNeuronLayer(FLayers[i]);
      For j := 0 to ALayer.NumNeurons -1 do
      begin
         ANeuron := ALayer.Neurons[j];
         For k := 0 to ANeuron.Weights.Count -1 do
         begin
            AWght := AWeights[c];
            Inc(c);
            ANeuron.Weights[k] := AWght;
         end;
      end;
   end;
end;

function TNeuralNet.Sigmoid(Activation, Response: Double): Double;
begin
   Result := 1.0 / (1.0 + exp(-Activation / response));
end;

procedure TNeuralNet.Update(Inputs, Outputs: TVector);
var
   cWeight: Integer;
   i, j, k: Integer;
   TmpInputs: TVector;
   NetInput: Double;
   ALayer: TNeuronLayer;
   ANeuron: TNeuron;
begin
   TmpInputs := TVector.Create;
   try
      TmpInputs.Assign(Inputs);
      if TmpInputs.Count <> FNumInputs then
      begin
         raise Exception.Create('Invalid number of inputs passed to TNeuralNet.Update!!!');
      end;

      For i := 0 to FNumHideLayers do
      begin
         if i > 0 then
         begin
            TmpInputs.Assign(Outputs);
         end;

         Outputs.Clear;
         cWeight := 0;

         ALayer := TNeuronLayer(FLayers[i]);
         For j := 0 to ALayer.NumNeurons -1 do
         begin
            ANeuron := ALayer.Neurons[j];
            NetInput := 0;

            For k := 0 to ANeuron.NumInputs -2 do
            begin
               NetInput := NetInput + (TmpInputs[cWeight] * ANeuron.Weights[k]);
               inc(cWeight);
            end;

            NetInput := NetInput + (ANeuron.Weights[ANeuron.NumInputs -1] * Params.Bias);

            NetInput := Sigmoid(NetInput, Params.ActivationResponse);
            Outputs.Add(NetInput);
            cWeight := 0;
         end;
      end;
   finally
      TmpInputs.Free;
   end;
end;

end.
